/**
     Program XCVNTest: Find the distribution of Cramer-von Mise test statistics

     Copyright (c) 2006 Yuanhui Xiao

     This program is free software; you can redistribute it 
     and/or modify it under the terms of the 
     GNU General Public License as published by 
     the Free Software Foundation;
     either version 2 of the License, 
     or (at your option) any later version.

     This program is distributed in the hope that it will be useful, 
     but WITHOUT ANY WARRANTY; without even the implied warranty of 
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See 
     the GNU General Public License for more details.

     You should have received a copy of the GNU General Public License 
     along with this program; if not, write to:

        the Free Software Foundation, Inc.
        51 Franklin St, Fifth
        Floor, Boston, MA 02110-1301, USA.

     Contact information: 
     E-mail: matyxx@langate.gsu.edu
     Adress:
        Department of Mathematics and Statistics
        Georgia State University
        750 COE, 30 Pryor Street
        Atlanta, GA 30303, USA
*/

#include "XCVMStat.h"
#include "XCVMDistEQ.h"
#include "XCVMDistNE1.h"
#include "XCVMDistNE2.h"
#include "XNumberCheckers.h"

#include <iostream>
#include <fstream>
#include <iomanip>
#include <iterator>
#include <ctime>

using namespace std;

void PrintDistFunct(int m, int n); 
      // print the entire distribution function for CVM test
      // m and n are sample sizes
void ComputePValues(int m, int n, int p, char** stats, int iFF);
      // compute p-values for statistic values. m and n are
      // sample sizes. If iFF = FFS, then convolutions are used
      // when computing p-values; if iFF = DIST, then convolutions
      // are not to be used
void ComputePValues(int m, int n, char* StatFileName, int iFF);
      // computing p-values for statistic values in the file
      // StatFileName. m, n and iFF are as above
void ComputeSampleStatPValue(const char* DataFileName, int iFF);
      // computing CVM statistic values and the corresponding 
      // p-values for a set of samples in the file DataFileName.
      // m, n and iFF are as above.
      
main(int argc, char** argv)
{
   if(argc == 1){ // no argument given, print usage of the package

     cout<<_NEWLINE
         <<">>>>>Usage: "<<_NEWLINE            
	 <<"     Enter: XCVMTest m n "<<_NEWLINE
	 <<"       Output: probability distribution of the Cramer-von Mises two-sample "
         <<_NEWLINE
         <<"               test statistic for the given sample sizes m and n. "
         <<_NEWLINE<<_NEWLINE			
         <<"     Enter: XCVMTest [-d|-f] m n t1 t2 ... tr "
         <<_NEWLINE
         <<"       Output: the p-values for the given statistic values t1, t2, ..., tr."           
         <<_NEWLINE<<_NEWLINE
         <<"     Enter: XCVMTest [-d|-f] m n StatFileName "
         <<_NEWLINE
         <<"       Output: the p-values for the given statistic values contained "
         <<_NEWLINE
         <<"               in the file StatFileName. "
         <<_NEWLINE<<_NEWLINE
         <<"     Enter: XCVMTest [-d|-f] DataFileName "
         <<_NEWLINE
         <<"       Output: Test statistic values and corresponding p-values "
         <<_NEWLINE
         <<"               for the pairs of samples contained in the file DatFileName. "
         <<_NEWLINE<<_NEWLINE			
	 <<"     Option \"-d\": p-values are computed via the distribution function; "
         <<_NEWLINE
	 <<"     Option \"-f\": p-values are computed by convoluting smaller frequency tables. "
         <<_NEWLINE
	 <<"     Default option is \"-f\", which consumes less memory. "
         <<_NEWLINE;
            
	 return 0;
   }

   cout<<_NEWLINE<<" You have entered: ";
   for(int kk=0; kk<argc; kk++)
      cout<<argv[kk]<<"  ";
   cout<<_NEWLINE<<_NEWLINE;
   
   int iFF = FFS; //iFF is one of the two: DIST, FFS
   int iOffSet = 0;

   if(argv[1][0] == '-'){ // determine the value of iFF

     switch(argv[1][1]){

        case 'd':
	  iFF = DIST;
	  break;

	case 'f':
	  iFF = FFS;
	  break;

        default:

          cerr<<"Input error! Please type the command \"XCVMTest\" for the usage of the program. "<<endl; 
	  return -1;
          break;
     }

     iOffSet = 1;
   }

   clock_t tbeg = clock();
   time_t  ctbeg = time(NULL);

   switch( argc-iOffSet )
   {
      case 2:  // compute stat and p-value for samples in a file

        try{

           ComputeSampleStatPValue( argv[1+iOffSet], iFF );

        }catch(int e){

           cerr<<"Input error! Please type the command \"XCVMTest\" for the usage of the program. "<<endl;
           return -1;

        }catch( ... ){

           cerr<<" Not enough memory for these m and n. "
               <<" Try the option -f if the option -d is used. "
               <<_NEWLINE;   
           return -2;
        }
        break;

      case 3:  // print entire distributions		

        // check part of the input

        if(iOffSet == 1 || !IsUnsigned(argv[1+iOffSet]) || !IsUnsigned(argv[2+iOffSet])){

           cerr<<"Input error! Please type the command \"XCVMTest\" for the usage of the program. "<<endl;
           return -1;
        }

        try{

            PrintDistFunct( atoi(argv[1+iOffSet]), atoi(argv[2+iOffSet]) ); 

        }catch(int e){

           cerr<<"Input error! Please type the command \"XCVMTest\" for the usage of the program. "<<endl;
           return -1;

        }catch( ... ){

           cerr<<"Not enough memory for these m and n. "
               <<_NEWLINE;
           return -2;
        }

        break;	   

      case 4:

        // check part of the input

        if(!IsUnsigned(argv[1+iOffSet]) || !IsUnsigned(argv[2+iOffSet])){

           cerr<<"Input error! Please type the command \"XCVMTest\" for the usage of the program. "<<endl;
           return -1;
        }
       
        try{

          if(IsDouble(argv[3+iOffSet]))
             ComputePValues( atoi(argv[1+iOffSet]), atoi(argv[2+iOffSet]), 1, &argv[3+iOffSet], iFF );   
          else            
             ComputePValues( atoi(argv[1+iOffSet]), atoi(argv[2+iOffSet]), argv[3+iOffSet], iFF );                                                         

        }catch(int e){
        
           cerr<<"Input error! Please type the command \"XCVMTest\" for the usage of the program. "<<endl;
           return -1;
       
        }catch( ... ){

            cerr<<" Not enough memory for these m and n. "
                <<" Try the option -f if the option -d is used. "
                <<_NEWLINE;

            return -2;
        }
        break;

      default: // compute p-values for given stats

        // check input

        if( !IsUnsigned(argv[1+iOffSet]) || !IsUnsigned(argv[2+iOffSet]) ){

           cerr<<"Input error! Please type the command \"XCVMTest\" for the usage of the program. "<<endl;
           return -1;                       
        }
        
        try{

           ComputePValues( atoi(argv[1+iOffSet]), atoi(argv[2+iOffSet]), argc-iOffSet-3, &argv[3+iOffSet], iFF);       

        }catch(int e){

           cerr<<"Input error! Please type the command \"XCVMTest\" for the usage of the program. "<<endl;
           return -1;

        }catch(...){

           cerr<<" Not enough memory for these m and n. "
               <<" Please try the option -f if the option -d is used. "
               <<_NEWLINE;

           return -5;
        }
        break;
   }

   clock_t tend = clock();
   time_t  ctend = time(NULL);

   cout<<setw(0)<<left
       <<_NEWLINE
       <<"The program started: "<<ctime(&ctbeg)
       <<"          and ended: "<<ctime(&ctend)
       <<"The computing time is ";

   if(tend > tbeg) 
      cout<<setprecision(4)<<( (double)(tend-tbeg)/CLOCKS_PER_SEC )<<" seconds. "
          <<_NEWLINE;
   else
      cout<<" approximately "<<difftime(ctend, ctbeg)<<" seconds. "
          <<_NEWLINE;
 
   return 0;
}

void PrintDistFunct(int m, int n)
{
   CXCVMDist* pDist;
   if(m == n){
       pDist = new CXCVMDistEQ(m, DIST);
   }else if( 0 == (m+n) % 2 ){
       pDist = new CXCVMDistNE1(m, n, DIST);
   }else{
       pDist = new CXCVMDistNE2(m, n, DIST);
   }
       
   pDist->PrintDistFunct( cout );

   delete pDist;
}

void ComputePValues(int m, int n, int p, char** stats, int iFF)
{
    // check arguments

    for(int k=0; k<p; k++)
       if(!IsDouble(stats[k]))
           throw -2;

    // computing

    CXCVMDist *pCVMDist;
    if(m == n){
       pCVMDist = new CXCVMDistEQ(m, iFF);
    }else if( 0 == (m+n) % 2 ){
       pCVMDist = new CXCVMDistNE1(m, n, iFF);
    }else{
       pCVMDist = new CXCVMDistNE2(m, n, iFF);
    }

    cout<<" m = "<<m<<", n = "<<n<<endl
        <<"           "<<"\t"
        <<" Statistic  "<<"\t"
        <<"   p-value    "<<"\t"
        <<"   p-value    "<<endl;

    cout<<right;
    for(int i=0; i<p; i++){

       double dStat = atof(stats[i]);
       double dPVal = pCVMDist->GetTailProb( dStat );
        
       cout<<"No. "<<setw(4)<<(i+1)<<".\t"
           <<setw(12)<<setprecision(8)<<fixed<<dStat<<"\t"
           <<setw(14)<<setprecision(8)<<scientific<<dPVal<<"\t"                
           <<setw(4)<<setprecision(12)<<fixed<<dPVal
           <<dPVal<<_NEWLINE;
    }

    cout<<_NEWLINE<<" *** End of program *** "<<_NEWLINE;

    delete pCVMDist;
}

void ComputePValues(int m, int n, char*  StatFileName, int iFF)
{
  // create the distribution function or frequency tables

   cout<<"Create the distribution function. please wait ... "
       <<_NEWLINE;

   CXCVMDist *pCVMDist;
   if(m == n){
       pCVMDist = new CXCVMDistEQ(m, iFF);
   }else if( 0 == (m+n) % 2 ){
       pCVMDist = new CXCVMDistNE1(m, n, iFF);
   }else{
       pCVMDist = new CXCVMDistNE2(m, n, iFF);
   }

   // open the data file

   ifstream is(StatFileName);
   if(!is){

      is.close();
      throw -3; // no file
   }
     
   // the results will be saved in the file [StatFileName].cvm and 
   // displayed in screen
   char* OutputFile = new char[5+strlen(StatFileName)];
   strcpy(OutputFile, StatFileName);
   strcat(OutputFile, ".cvm");

   ofstream os(OutputFile);

   // read statistic values and calculate the p-values.
   
   double dStat, dPValue; //

   int nCount;  // a counter, counting the number of statistics 

   cout<<" m = "<<m<<", n = "<<n<<_NEWLINE
       <<"           "<<"\t"
       <<"Statistic"<<"\t"
       <<"   p-value    "<<"\t"
       <<"   p-value    "<<_NEWLINE;

     os<<" m = "<<m<<", n = "<<n<<_NEWLINE
       <<"           "<<"\t"
       <<"Statistic"<<"\t"
       <<"   p-value    "<<"\t"
       <<"   p-value    "<<_NEWLINE;

   istream_iterator<double> readin(is);

   nCount = 0;   
  
   cout<<right;
      os<<right;
   while( !is.eof() ){

      dStat = *readin;
      dPValue = pCVMDist->GetTailProb( dStat );	
      nCount++;
	  
      cout<<"No. "<<setw(4)<<nCount<<".\t"
          <<setw(10)<<setprecision(6)<<fixed<<dStat<<"\t"
          <<setw(14)<<setprecision(8)<<scientific<<dPValue<<"\t"
          <<setw(14)<<setprecision(12)<<fixed<<dPValue<<_NEWLINE;
	  
        os<<"No. "<<setw(4)<<right<<nCount<<".\t"
          <<setw(10)<<setprecision(6)<<fixed<<dStat<<"\t"
          <<setw(14)<<setprecision(8)<<scientific<<dPValue<<"\t"
          <<setw(14)<<setprecision(12)<<fixed<<dPValue<<_NEWLINE;
	  
      ++readin;
   } 
   os<<_NEWLINE<<" *** End of file *** "<<_NEWLINE;

   is.close(); // close the data file
   os.close(); // close the result file

   cout<<"\n***The results have been saved into the file \""
       <<StatFileName<<".cvm\". "
       <<_NEWLINE;

   delete pCVMDist;
}

void ComputeSampleStatPValue(const char* DataFileName, int iFF) 
{
    int m, n, i;
    double dStat, dPVal;
 
    ifstream is(DataFileName);
    if(!is){

       is.close();
       throw -3;  // no such file
    }

    char* OutputFile = new char[5+strlen(DataFileName)];
    strcpy(OutputFile, DataFileName);
    strcat(OutputFile, ".cvm");
    ofstream os(OutputFile);

    is>>m>>n;

    CXCVMStat *pCVMStat = new CXCVMStat(m, n);

    CXCVMDist *pCVMDist;
    if(m == n){
       pCVMDist = new CXCVMDistEQ(m, iFF);
    }else if( 0 == (m+n) % 2 ){
       pCVMDist = new CXCVMDistNE1(m, n, iFF);
    }else{
       pCVMDist = new CXCVMDistNE2(m, n, iFF);
    }

    double *z = new double[m+n];
    z--; // so that x is one-based

    cout<<"  m = "<<m<<", n = "<<n<<_NEWLINE<<_NEWLINE
        <<"            "<<"  statistic "<<"\t"
                        <<"   p-value   "<<"\t"
                        <<"   p-value   "<<_NEWLINE;

      os<<"  m = "<<m<<", n = "<<n<<_NEWLINE<<_NEWLINE
        <<"              "<<"  statistic "<<"\t"
                          <<"   p-value   "<<"\t"
                          <<"   p-vlaue   "<<_NEWLINE;

   int nCount=0;
   for(i=1; i<=m+n; i++) is>>z[i]; // read in first sample

   cout<<right;
     os<<right;
   do{

      nCount++;

      dStat = pCVMStat->ComputeStatistic( z );
      dPVal = pCVMDist->GetTailProb(dStat);
	           
      cout<<"No. "<<setw(4)<<nCount<<".\t"
          <<setw(12)<<setprecision(8)<<fixed<<dStat<<"\t"
          <<setw(14)<<setprecision(12)<<fixed<<dPVal<<"\t"
          <<setw(14)<<setprecision(8)<<scientific<<dPVal<<_NEWLINE;

        os<<"No. "<<setw(4)<<nCount<<".\t"
          <<setw(12)<<setprecision(8)<<fixed<<dStat<<"\t"
          <<setw(14)<<setprecision(12)<<fixed<<dPVal<<"\t"
          <<setw(14)<<setprecision(8)<<scientific<<dPVal<<_NEWLINE;
	
      for(i=1; i<=m+n; i++) is>>z[i]; // read in next sample	

   }while(!is.eof());


      os<<_NEWLINE<<" *** End of file *** "<<_NEWLINE;
    cout<<_NEWLINE<<" *** End of file *** "<<_NEWLINE;

    os.close();
    is.close();	

    delete [] ++z;
    delete pCVMStat;
    delete pCVMDist;

    cout<<_NEWLINE
        <<"The result is stored in file "<<DataFileName
        <<".cvm"<<_NEWLINE;
}

